/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package org.apithefire.servlet.embedded;

import java.io.IOException;
import java.util.Properties;

import org.apithefire.util.io.RuntimeIoException;
import org.apithefire.util.lang.Objects;

/**
 * A default content type resolver implementation.
 * 
 * <p>
 * This implementation reads content type mappings from
 * {@code DefaultContentTypes.properties} relative to this class.
 * </p>
 * 
 * @author <a href="mailto:bachew@gmail.com">Chew Boon Aik</a>
 */
public class DefaultContentTypeResolver implements ContentTypeResolver {
	
	public static final String APPLICATION_OCTET_STREAM = "application/octet-stream";
	
	private final Properties contentTypes;
	
	/**
	 * Constructs a new instance.
	 */
	public DefaultContentTypeResolver() {
		try {
			contentTypes = new Properties();
			contentTypes.load(getClass().getResourceAsStream(
					"DefaultContentTypes.properties"));
		} catch (IOException e) {
			throw new RuntimeIoException(e);
		}
	}

	/**
	 * {@inheritDoc}. If the specified suffix is not recognized,
	 * {@link #APPLICATION_OCTET_STREAM application/octet-stream} will be
	 * returned.
	 */
	public String resolveContentType(String suffix) {
		String contentType = (String) contentTypes.get(Objects.nonNull(suffix));
		
		return contentType == null ? APPLICATION_OCTET_STREAM : contentType;
	}

}
