/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package org.apithefire.servlet.embedded;

import java.io.IOException;
import java.util.List;
import java.util.ListIterator;

import javax.servlet.Filter;
import javax.servlet.FilterChain;
import javax.servlet.Servlet;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;

import org.apithefire.util.lang.Objects;
import org.apithefire.util.lang.Path;

class EmbeddedUtil {

	private static final FilterChain EMPTY_CHAIN = new FilterChain() {
		public void doFilter(ServletRequest request,
				ServletResponse response) throws IOException,
				ServletException {
		}
	};
	
	private static class ServletFilterChain implements FilterChain {
		
		private final Servlet servlet;
		
		public ServletFilterChain(Servlet servlet) {
			this.servlet = Objects.nonNull(servlet);
		}

		public void doFilter(ServletRequest request, ServletResponse response)
				throws IOException, ServletException {
			servlet.service(request, response);
		}
		
	}
	
	private static class FilterFilterChain implements FilterChain {
		
		private final Filter filter;
		private final FilterChain filterChain;
		
		public FilterFilterChain(Filter filter, FilterChain filterChain) {
			this.filter = Objects.nonNull(filter);
			this.filterChain = Objects.nonNull(filterChain);
		}

		public void doFilter(ServletRequest request,
				ServletResponse response) throws IOException,
				ServletException {
			filter.doFilter(request, response, filterChain);
		}
		
	}
	
	private EmbeddedUtil() {		
	}
	
	public static FilterChain chain(Servlet servlet, List<Filter> filters) {
		FilterChain chain = EMPTY_CHAIN;
		
		if (servlet != null) {
			chain = new ServletFilterChain(servlet);
		}
		if (!filters.isEmpty()) {
			ListIterator<Filter> filterIter = filters.listIterator(filters.size());
			
			while (filterIter.hasPrevious()) {
				Filter filter = filterIter.previous();
				chain = new FilterFilterChain(filter, chain);
			}
		}
		return chain;
	}
	
	public static Path splitRequestPath(String path) {
		if (path == null) {
			return Path.EMPTY_PATH;
		}
		return Path.split(path).trim();
	}

	public static Path splitLastComponentAsSuffix(Path path) {
		if (path.isEmpty()) {
			return Path.EMPTY_PATH;
		}
		
		Path lastComp = Path.split(path.componentAt(path.length() - 1), ".");
		
		if (lastComp.length() < 2) {
			return Path.EMPTY_PATH;
		}
		return lastComp.subPath(1);
	}
	
	public static String parseSimpleSuffix(Path path) {
		if (path.isEmpty()) {
			return "";
		}
		
		String lastComp = path.componentAt(path.length() - 1);
		int lastDotIndex = lastComp.lastIndexOf('.');
		
		if (lastDotIndex == -1) {
			return "";
		}
		return lastComp.substring(lastDotIndex + 1);
	}

}
