/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package org.apithefire.servlet.embedded;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import javax.servlet.Filter;

import org.apithefire.util.lang.Objects;
import org.apithefire.util.lang.Path;

class FilterPathMap {

	private static abstract class FilterHolder {
		
		private Filter filter;
		
		public FilterHolder(Filter filter) {
			this.filter = Objects.nonNull(filter);
		}
		
		public Filter getFilter() {
			return filter;
		}
		
		public abstract boolean matches(Path path, Path suffix);
		
	}
	
	private static class ExactFilterHolder extends FilterHolder {
		
		private Path exactPath;
		
		public ExactFilterHolder(Path exactPath, Filter filter) {
			super(filter);
			this.exactPath = Objects.nonNull(exactPath);
		}
		
		@Override
		public boolean matches(Path path, Path suffix) {
			return exactPath.equals(path);
		}
		
	}
	
	private static class PrefixFilterHolder extends FilterHolder {
		
		private Path prefix;
		
		public PrefixFilterHolder(Path prefix, Filter filter) {
			super(filter);
			this.prefix = Objects.nonNull(prefix);
		}

		@Override
		public boolean matches(Path path, Path suffix) {
			return path.startsWith(prefix);
		}
		
	}
	
	private static class SuffixFilterHolder extends FilterHolder {
		
		private Path suffix;
		
		public SuffixFilterHolder(Path suffix, Filter filter) {
			super(filter);
			this.suffix = Objects.nonNull(suffix);
		}

		@Override
		public boolean matches(Path path, Path suffix) {
			return suffix.endsWith(this.suffix);
		}
		
	}
	
	private final List<FilterHolder> filterHolders;
	
	public FilterPathMap() {
		filterHolders = new ArrayList<FilterHolder>();
	}
	
	public List<Filter> getMatches(Path path) {
		List<Filter> filters = new ArrayList<Filter>(filterHolders.size());
		Path suffix = EmbeddedUtil.splitLastComponentAsSuffix(path);
		
		for (FilterHolder filterHolder : filterHolders) {
			if (filterHolder.matches(path, suffix)) {
				filters.add(filterHolder.getFilter());
			}
		}
		return filters;
	}
	
	public void mapExact(Path exactPath, Filter filter) {
		filterHolders.add(new ExactFilterHolder(exactPath, filter));
	}
	
	public void mapPrefix(Path prefix, Filter filter) {
		filterHolders.add(new PrefixFilterHolder(prefix, filter));
	}
	
	public void mapSuffix(Path suffix, Filter filter) {
		filterHolders.add(new SuffixFilterHolder(suffix, filter));
	}
	
	public Set<Filter> getFilters() {
		Set<Filter> filters = new HashSet<Filter>();
		
		for (FilterHolder filterHolder : filterHolders) {
			filters.add(filterHolder.getFilter());
		}
		return filters;
	}
	
	@Override
	public FilterPathMap clone() {
		FilterPathMap clone = new FilterPathMap();
		
		clone.filterHolders.addAll(filterHolders);
		return clone;
	}
	
}
