/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package org.apithefire.servlet.embedded;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import javax.servlet.Filter;
import javax.servlet.Servlet;
import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;

import org.apithefire.util.lang.Builder;
import org.apithefire.util.lang.BuilderFactory;
import org.apithefire.util.lang.Objects;

/**
 * A servlet with filtering {@linkplain Filter filters}.
 * <p>
 * TODO: link to an example.
 * </p>
 * 
 * @author <a href="mailto:bachew@gmail.com">Chew Boon Aik</a>
 */
public class FilteredServlet implements Servlet {
	
	/**
	 * A {@link FilteredServlet} builder.
	 * 
	 * @author <a href="mailto:bachew@gmail.com">Chew Boon Aik</a>
	 */
	public static class FilteredServletBuilder implements Builder {
		
		private Servlet servlet;
		private final List<Filter> filters = new ArrayList<Filter>();
		
		private FilteredServletBuilder() {
		}
		
		/**
		 * Sets the specified main servlet.
		 * 
		 * @return this builder
		 */
		public FilteredServletBuilder setServlet(Servlet servlet) {
			this.servlet = Objects.nonNull(servlet);
			return this;
		}
		
		/**
		 * Adds the specified filter.
		 * 
		 * @return this builder
		 */
		public FilteredServletBuilder addFilter(Filter filter) {
			this.filters.add(Objects.nonNull(filter));
			return this;
		}
		
		/**
		 * Builds a new {@link FilteredServlet}.
		 * 
		 * @return the built {@link FilteredServlet}
		 */
		public FilteredServlet build() {
			FilteredServlet filteredServlet = new FilteredServlet();
			
			filteredServlet.container = new ServletAndFilterContainer();
			
			if (servlet != null) {
				filteredServlet.servlet = servlet;
				filteredServlet.container.add(servlet);
			}
			filteredServlet.filters = new ArrayList<Filter>(filters);
			filteredServlet.container.addAll(filters);
			return filteredServlet;
		}
		
	}
	
	private Servlet servlet;
	private List<Filter> filters;
	private ServletAndFilterContainer container;
	private ServletConfig servletConfig;
	
	private FilteredServlet() {
	}
	
	/**
	 * Returns a new {@link FilteredServletBuilder} builder.
	 */
	@BuilderFactory
	public static FilteredServletBuilder newBuilder() {
		return new FilteredServletBuilder();
	}

	public void init(ServletConfig servletConfig) throws ServletException {
		this.servletConfig = servletConfig;
		container.init(servletConfig.getServletContext());
	}

	public void destroy() {
		container.destroy();
	}

	public ServletConfig getServletConfig() {
		return servletConfig;
	}

	public String getServletInfo() {
		return toString();
	}

	public void service(ServletRequest request, ServletResponse response)
			throws ServletException, IOException {
		EmbeddedUtil.chain(servlet, filters).doFilter(request, response);
	}

}
