/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package org.apithefire.servlet.embedded;

import java.io.IOException;
import java.util.List;

import javax.servlet.Filter;
import javax.servlet.FilterChain;
import javax.servlet.Servlet;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletRequestWrapper;
import javax.servlet.http.HttpServletResponse;

import org.apithefire.servlet.embedded.ServletPathMap.Match;
import org.apithefire.util.lang.BuilderFactory;
import org.apithefire.util.lang.Objects;
import org.apithefire.util.lang.Path;

/**
 * A servlet and filter path mapping servlet.
 * 
 * <p>
 * {@code PathMapServlet} is written to replaced static {@code web.xml}
 * configuration with programmatic configuration. {@code PathMapServlet} does
 * not attempt to be fully comply with servlet specification because web
 * applications should not rely too much on {@code web.xml}.
 * </p>
 * 
 * TODO: link to an example
 * 
 * @author <a href="mailto:bachew@gmail.com">Chew Boon Aik</a>
 */
public class PathMapServlet extends HttpServlet {
	
	/**
	 * A {@link PathMapServlet} builder.
	 * 
	 * @author <a href="mailto:bachew@gmail.com">Chew Boon Aik</a>
	 */
	public static class PathMapServletBuilder {
		
		private final ServletPathMap servletMap = new ServletPathMap();
		private final FilterPathMap filterMap = new FilterPathMap();
		
		private PathMapServletBuilder() {
		}
		
		/**
		 * Maps an exact path servlet.
		 * <p>
		 * Assuming {@code exactPath = new Path("exact", "path")}. In
		 * {@code web.xml}, it would look like
		 * </p>
		 * 
		 * <pre>
		 * &lt;servlet-mapping&gt;
		 *   &lt;servlet-name&gt;servlet&lt;/servlet-name&gt;
		 *   &lt;url-pattern&gt;/exact/path&lt;/url-pattern&gt;
		 * &lt;/servlet-mapping&gt;
		 * </pre>
		 * 
		 * @return this builder
		 */
		public PathMapServletBuilder mapExact(Path exactPath, Servlet servlet) {
			servletMap.mapExact(exactPath, servlet);
			return this;
		}
		
		/**
		 * Maps a prefix servlet.
		 * <p>
		 * Assuming {@code prefix = new Path("prefix", "path")}. In
		 * {@code web.xml}, it would look like
		 * </p>
		 * 
		 * <pre>
		 * &lt;servlet-mapping&gt;
		 *   &lt;servlet-name&gt;servlet&lt;/servlet-name&gt;
		 *   &lt;url-pattern&gt;/prefix/path/*&lt;/url-pattern&gt;
		 * &lt;/servlet-mapping&gt;
		 * </pre>
		 * 
		 * @return this builder
		 */
		public PathMapServletBuilder mapPrefix(Path prefix, Servlet servlet) {
			servletMap.mapPrefix(prefix, servlet);
			return this;
		}

		
		/**
		 * Maps a prefix servlet.
		 * <p>
		 * Assuming {@code suffix = new Path("suf", "fix")}. In
		 * {@code web.xml}, it would look like
		 * </p>
		 * 
		 * <pre>
		 * &lt;servlet-mapping&gt;
		 *   &lt;servlet-name&gt;servlet&lt;/servlet-name&gt;
		 *   &lt;url-pattern&gt;*.suf.fix&lt;/url-pattern&gt;
		 * &lt;/servlet-mapping&gt;
		 * </pre>
		 * 
		 * @return this builder
		 */
		public PathMapServletBuilder mapSuffix(Path suffix, Servlet servlet) {
			servletMap.mapSuffix(suffix, servlet);
			return this;
		}
		
		/**
		 * Maps the specified default servlet.
		 * 
		 * @return this builder
		 */
		public PathMapServletBuilder mapDefault(Servlet servlet) {
			servletMap.mapDefault(servlet);
			return this;
		}
		
		/**
		 * Maps an exact path filter.
		 * 
		 * @see #mapExact(Path, Servlet)
		 * @return this builder
		 */
		public PathMapServletBuilder mapExact(Path exactPath, Filter filter) {
			filterMap.mapExact(exactPath, filter);
			return this;
		}
		
		/**
		 * Maps a prefix path filter.
		 * 
		 * @see #mapPrefix(Path, Servlet)
		 * @return this builder
		 */
		public PathMapServletBuilder mapPrefix(Path prefix, Filter filter) {
			filterMap.mapPrefix(prefix, filter);
			return this;
		}
		
		/**
		 * Maps a suffix filter.
		 * 
		 * @see #mapSuffix(Path, Servlet)
		 * @return this builder
		 */
		public PathMapServletBuilder mapSuffix(Path suffix, Filter filter) {
			filterMap.mapSuffix(suffix, filter);
			return this;
		}
		
		/**
		 * Builds a new {@code PathMapServlet}.
		 * 
		 * @return the built {@code PathMapServlet}
		 */
		public PathMapServlet build() {
			PathMapServlet pathMapServlet = new PathMapServlet();
			
			pathMapServlet.servletMap = servletMap.clone();
			pathMapServlet.filterMap = filterMap.clone();
			pathMapServlet.container = new ServletAndFilterContainer();
			pathMapServlet.container.addAll(servletMap.getServlets());
			pathMapServlet.container.addAll(filterMap.getFilters());
			return pathMapServlet;
		}
		
	}
	
	private static class ServletMatchRequest extends
			HttpServletRequestWrapper {
		
		private final Match servletMatch;

		public ServletMatchRequest(Match servletMatch,
				HttpServletRequest request) {
			super(request);
			this.servletMatch = Objects.nonNull(servletMatch);
		}

		@Override
		public String getServletPath() {
			Path servletPath = servletMatch.getServletPath();
			
			return "/" + servletPath.trim().toString();
		}

		@Override
		public String getPathInfo() {
			Path pathInfo = servletMatch.getPathInfo();
			
			if (pathInfo.isEmpty()) {
				return null;
			}
			return "/" + pathInfo.trim().toString();
		}
		
	}
	
	private ServletPathMap servletMap;
	private FilterPathMap filterMap;
	private ServletAndFilterContainer container;
	
	private PathMapServlet() {
	}
	
	/**
	 * Returns a new {@link PathMapServletBuilder}.
	 */
	@BuilderFactory
	public static PathMapServletBuilder newBuilder() {
		return new PathMapServletBuilder();
	}

	/**
	 * Initializes all mapped servlets and filters.
	 */
	@Override
	public void init() throws ServletException {
		container.init(getServletConfig().getServletContext());
	}

	/**
	 * Destroys all mapped servlets and filters.
	 */
	@Override
	public void destroy() {
		container.destroy();
	}
	
	/**
	 * Delegates request to appropriate servlet and filtered accordingly.
	 */
	@Override
	protected void service(HttpServletRequest request,
			HttpServletResponse response) throws ServletException, IOException {
		Path path = EmbeddedUtil.splitRequestPath(request.getPathInfo());
		Match servletMatch = servletMap.getMatch(path);
		
		Servlet servlet = null;
		List<Filter> filters = filterMap.getMatches(path);
		
		if (servletMatch != null) {
			servlet = servletMatch.getServlet();
		}
		
		FilterChain chain = EmbeddedUtil.chain(servlet, filters);
		
		if (servletMatch != null) {
			request = new ServletMatchRequest(
					servletMatch, request);
		}
		chain.doFilter(request, response);
	}

}
