/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package org.apithefire.servlet.embedded;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.Iterator;
import java.util.List;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apithefire.servlet.RuntimeServletException;
import org.apithefire.util.lang.BuilderFactory;
import org.apithefire.util.lang.Objects;
import org.apithefire.util.lang.Path;
import org.apithefire.util.net.URIs;
import org.apithefire.util.resource.Resource;
import org.apithefire.util.resource.ResourceNotFoundException;

/**
 * A resource servlet.
 * <p>
 * A {@link ResourceServlet} serves static resource typically reside in
 * classpath or filesystem.
 * </p>
 * 
 * @author <a href="mailto:bachew@gmail.com">Chew Boon Aik</a>
 */
public class ResourceServlet extends HttpServlet {

	/**
	 * Default buffer size, {@value}.
	 */
	public static final int DEFAULT_BUFFER_SIZE = 1024;
	
	/**
	 * Default index file name, {@value}.
	 */
	public static final String DEFAULT_INDEX_FILE_NAME = "index.html";
	
	/**
	 * Enable directory listing by default? {@value}.
	 */
	public static final boolean DEFAULT_DIRECTORY_LISTING = false;
	
	/**
	 * Default base resource name displayed in directory listing.
	 */
	public static final String DEFAULT_BASE_RESOURCE_NAME = "Resources";
	
	/**
	 * A {@link ResourceServlet} builder.
	 * 
	 * @author <a href="mailto:bachew@gmail.com">Chew Boon Aik</a>
	 */
	public static class ResourceServletBuilder {
		
		private Resource baseResource;
		private int bufferSize = DEFAULT_BUFFER_SIZE;
		private ContentTypeResolver contentTypeResolver;
		private String indexFileName = DEFAULT_INDEX_FILE_NAME;
		private boolean directoryListing = DEFAULT_DIRECTORY_LISTING;
		private String baseResourceName = DEFAULT_BASE_RESOURCE_NAME;
		
		private ResourceServletBuilder() {
		}
		
		/**
		 * Sets the specified base resource.
		 * 
		 * @return this builder
		 */
		public ResourceServletBuilder setBaseResource(Resource baseResource) {
			this.baseResource = Objects.nonNull(baseResource);
			return this;
		}

		/**
		 * Sets the specified buffer size in bytes used to transfer data from
		 * resource to servlet output.
		 * 
		 * @see ResourceServlet#DEFAULT_BUFFER_SIZE
		 * @return this builder
		 */
		public ResourceServletBuilder setBufferSize(int bufferSize) {
			if (bufferSize < 1) {
				throw new IllegalArgumentException("Buffer size is less than 1.");
			}
			this.bufferSize = bufferSize;
			return this;
		}
		
		/**
		 * Sets the specified {@link ContentTypeResolver}.
		 * 
		 * @return this builder
		 */
		public ResourceServletBuilder setContentTypeResolver(
				ContentTypeResolver contentTypeResolver) {
			this.contentTypeResolver = Objects.nonNull(contentTypeResolver);
			return this;
		}
		
		/**
		 * Sets the specified index file name.
		 * 
		 * @see ResourceServlet#DEFAULT_INDEX_FILE_NAME
		 * @return this builder
		 */
		public ResourceServletBuilder setIndexFileName(String indexFileName) {
			this.indexFileName = indexFileName;
			return this;
		}
		
		/**
		 * Enables or disables directory listing.
		 * 
		 * @see ResourceServlet#DEFAULT_DIRECTORY_LISTING
		 * @return this builder
		 */
		public ResourceServletBuilder setDirectoryListing(boolean directoryListing) {
			this.directoryListing = directoryListing;
			return this;
		}
		
		/**
		 * Sets base resource name displayed in directory listing.
		 * 
		 * @return this builder
		 */
		public ResourceServletBuilder setBaseResourceName(String baseResourceName) {
			this.baseResourceName = Objects.nonNull(baseResourceName);
			return this;
		}
		
		/**
		 * Builds a {@link ResourceServlet}
		 * 
		 * @return the built {@link ResourceServlet}
		 * @throws RuntimeServletException if base resource is not set.
		 */
		public ResourceServlet build() {
			ResourceServlet resourceServlet = new ResourceServlet();

			if (baseResource == null) {
				throw new RuntimeServletException("Base resource not set.");
			}
			resourceServlet.baseResource = baseResource;
			resourceServlet.bufferSize = bufferSize;
			
			if (contentTypeResolver == null) {
				resourceServlet.contentTypeResolver = new DefaultContentTypeResolver();
			} else {
				resourceServlet.contentTypeResolver = contentTypeResolver;
			}
			resourceServlet.indexFileName = indexFileName;
			resourceServlet.directoryListing = directoryListing;
			resourceServlet.baseResourceName = baseResourceName;
			return resourceServlet;
		}
	}
	
	protected static final String HEADER_IF_MODIFIED_SINCE = "If-Modified-Since";
	protected static final String HEADER_LAST_MODIFIED = "Last-Modified";

	private Resource baseResource;
	private int bufferSize;
	private ContentTypeResolver contentTypeResolver;
	private String indexFileName;
	private boolean directoryListing;
	private String baseResourceName;
	private ThreadLocal<DateFormat> currentDateFormat = new ThreadLocal<DateFormat>() {
		@Override
		protected DateFormat initialValue() {
			return new SimpleDateFormat("dd-MMM-yyyy hh:mm");
		}
	};
	
	private ResourceServlet() {
	}
	
	/**
	 * Returns a new {@link ResourceServletBuilder}.
	 */
	@BuilderFactory
	public static ResourceServletBuilder newBuilder() {
		return new ResourceServletBuilder();
	}
	
	/**
	 * Returns base resource of this servlet.
	 */
	public Resource getBaseResource() {
		return baseResource;
	}

	/**
	 * Returns buffer size in bytes used to transfer data from resource to
	 * servlet output.
	 */
	public int getBufferSize() {
		return bufferSize;
	}

	/**
	 * Returns content type resolver of this servlet.
	 */
	public ContentTypeResolver getContentTypeResolver() {
		return contentTypeResolver;
	}
	
	/**
	 * Returns name used to resolve index file when a directory is accessed.
	 * 
	 * @see #DEFAULT_INDEX_FILE_NAME
	 */
	public String getIndexFileName() {
		return indexFileName;
	}
	
	/**
	 * Returns true if directory listing is enabled, false otherwise.
	 */
	public boolean getDirectoryListing() {
		return directoryListing;
	}
	
	/**
	 * Returns base resource name displayed in directory listing.  
	 */
	public String getBaseResourceName() {
		return baseResourceName;
	}

	/**
	 * Serves the appropriate resource.
	 */
	@Override
	protected void doGet(HttpServletRequest request, HttpServletResponse response)
			throws ServletException, IOException {
		
		String pathInfo = request.getPathInfo();
		
		if (pathInfo == null) {
			pathInfo = "/";
		}
		
		Path path = EmbeddedUtil.splitRequestPath(pathInfo);
		Resource resource = null;
		
		try {
			resource = baseResource.getResource(path);
		} catch (ResourceNotFoundException e) {
			response.sendError(HttpServletResponse.SC_NOT_FOUND);
			return;
		}
		
		boolean fromDirectory = false;
		
		if (resource.isDirectory()) {
			fromDirectory = true;
			
			if (!pathInfo.endsWith("/")) {
				response.sendRedirect(pathInfo + "/");
				return;
			}
			
			Path indexPath = null;
			Resource indexResource = null;
			
			if (indexFileName != null) {
				indexPath = path.concat(indexFileName);
				try {
					indexResource = baseResource.getResource(indexPath);
				} catch (ResourceNotFoundException e) {
				}
			}
			if (indexResource != null) {
				path = indexPath;
				resource = indexResource;
			} else if (directoryListing) {
				directoryListing(request, response, path, resource);
				return;
			} else {
				response.sendError(HttpServletResponse.SC_FORBIDDEN);
				return;
			}
		}
		if (!fromDirectory) {
			long ifModifiedSince = request.getDateHeader(HEADER_IF_MODIFIED_SINCE);
			long lastModified = -1;
			Date lastModifiedTime = resource.getFileModifiedTime();
			
			if (lastModifiedTime != null) {
				lastModified = lastModifiedTime.getTime() / 1000 * 1000;
				response.setDateHeader(HEADER_LAST_MODIFIED, lastModified);
			}
			if (ifModifiedSince != -1 && lastModified != -1) {
				if (lastModified <= ifModifiedSince) {
					response.sendError(HttpServletResponse.SC_NOT_MODIFIED);
					return;
				}
			}
		}

		String suffix = EmbeddedUtil.parseSimpleSuffix(path);
		String contentType = contentTypeResolver.resolveContentType(suffix);
		
		response.setContentType(contentType);
		
		OutputStream out = response.getOutputStream();
		InputStream content = resource.openFile();
		
		byte[] buffer = new byte[bufferSize];
		int n = content.read(buffer);
		
		while (n != -1) {
			out.write(buffer, 0, n);
			n = content.read(buffer);
		}
		response.flushBuffer();
	}
	
	private void directoryListing(HttpServletRequest request,
			HttpServletResponse response, Path path, Resource dir)
			throws IOException {
		response.setContentType("text/html");
		
		PrintWriter out = response.getWriter();
		
		path = new Path(baseResourceName).concat(path);
		
		out.println("<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN");
		out.println("    \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">");
		out.println("<html xmlns=\"http://www.w3.org/1999/xhtml\">");
		out.println("<head>");
		
		out.print("<title>");
		out.print(path.toString());
		out.println("</title>");
		
		out.println("<style type=\"text/css\">");
		out.println("/*<![CDATA[*/");
		out.println("h1 {");
		out.println("  font-size: 14pt;");
		out.println("}");
		out.println("tr {");
		out.println("  font-family: Courier New;");
		out.println("  font-size: 10pt;");
		out.println("}");
		out.println("tr.header {");
		out.println("  font-weight: bold;");
		out.println("}");
		out.println("td.separator {");
		out.println("  border-right: 2px dotted #000000;");
		out.println("}");
		out.println("/*]]>*/");
		out.println("</style>");
		
		out.println("</head>");
		out.println("<body>");
		
		out.print("<h1>");
		
		for (int i = 0; i < path.length(); ++i) {
			int backs = path.length() - (i + 1);
			
			if (i > 0) {
				out.print('/');
			}
			if (backs > 0) {
				out.print("<a href=\"..");
				
				for (int j = 1; j < backs; ++j) {
					out.print("/..");
				}
				out.print("\">");
			}
			out.print(path.componentAt(i));
			
			if (backs > 0) {
				out.print("</a>");
			}
		}
		out.println("</h1>");
		
		Collection<String> names = dir.list();
		Collection<String> dirNames = new ArrayList<String>(names.size());
		Collection<String> fileNames = new ArrayList<String>(names.size());
		List<Resource> files = new ArrayList<Resource>(names.size());
		
		for (String name : names) {
			Resource child = dir.getResource(new Path(name));
			
			if (child.isDirectory()) {
				dirNames.add(name);
			}
			if (child.isFile()) {
				fileNames.add(name);
				files.add(child);
			}
		}
		
		out.println("<table border=\"0\" width=\"100%\" summary=\"listing\">");
		out.print("<tr class=\"header\">");
		out.print("<td width=\"25%\">Directory");
		
		if (path.length() > 1) {
			out.print(" <a href=\"..\">[Up One Level]</a>");
		}
		out.print("</td>");
		out.print("<td width=\"2%\" class=\"separator\">&nbsp;</td>");
		out.print("<td width=\"3%\"></td>");
		out.print("<td width=\"40%\">File</td>");
		out.print("<td width=\"20%\">Last Modified</td>");
		out.print("<td width=\"10%\" align=\"right\">Size</td>");
		out.println("</tr>");
		
		Iterator<String> dirNameIter = dirNames.iterator();
		Iterator<String> fileNameIter = fileNames.iterator();
		Iterator<Resource> fileIter = files.iterator();
		DateFormat dateFormat = currentDateFormat.get();
		StringBuilder escapeHtmlTmpBuf = new StringBuilder();
		
		while (true) {
			boolean hasDir = dirNameIter.hasNext();
			boolean hasFile = (fileNameIter.hasNext() && fileIter.hasNext());
			
			if (hasDir || hasFile) {
				out.print("<tr>");
				
				if (hasDir) {
					String dirName = dirNameIter.next();
					
					out.print("<td><a href=\"");
					out.print(URIs.encode(dirName));
					out.print("/\">");
					out.print(escapeHtml(dirName, escapeHtmlTmpBuf));
					out.print("</a></td>");
				} else {
					out.print("<td></td>");
				}
				out.print("<td class=\"separator\">&nbsp;</td><td></td>");
				
				if (hasFile) {
					String fileName = fileNameIter.next();
					Resource file = fileIter.next();
					Date modifiedTime = file.getFileModifiedTime();
					long size = file.getFileSize();
					
					out.print("<td><a href=\"");
					out.print(URIs.encode(fileName));
					out.print("\">");
					out.print(escapeHtml(fileName, escapeHtmlTmpBuf));
					out.print("</a></td><td>");
					
					if (modifiedTime == null) {
						out.print("-");
					} else {
						out.print(dateFormat.format(modifiedTime));
					}
					out.print("</td><td align=\"right\">");
					
					if (size == -1) {
						out.print("-");
					} else {
						out.print(size);
					}
					out.print("</td>");
				} else {
					out.print("<td></td><td></td><td></td>");
				}
				out.println("</tr>");
			} else {
				break;
			}
		}
		out.println("</table>");
		
		out.println("</body>");
		out.println("</html>");
		response.flushBuffer();
	}
	
	private static String escapeHtml(String str, StringBuilder tmpBuf) {
		tmpBuf.setLength(0);
		
		for (int i = 0; i < str.length(); ++i) {
			char ch = str.charAt(i);
			
			if (ch == '&') {
				tmpBuf.append("&amp;");
			} else if (ch == ' ') {
				tmpBuf.append("&nbsp;");
			} else if (ch == '<') {
				tmpBuf.append("&lt;");
			} else if (ch == '>') {
				tmpBuf.append("&gt;");
			} else {
				tmpBuf.append(ch);
			}
		}
		return tmpBuf.toString();
	}
	
}
