/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package org.apithefire.servlet.embedded;

import java.util.Collections;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Map;
import java.util.TreeMap;

import javax.servlet.FilterConfig;
import javax.servlet.ServletConfig;
import javax.servlet.ServletContext;

import org.apithefire.util.lang.Objects;

/**
 * A combination of {@link ServletConfig} and {@link FilterConfig}.
 * 
 * @author <a href="mailto:bachew@gmail.com">Chew Boon Aik</a>
 */
public class ServletAndFilterConfig implements ServletConfig, FilterConfig {
	
	private final ServletContext servletContext;
	private final String name;
	private final Map<String, String> initParamMap;
	
	/**
	 * Constructs a new instance.
	 */
	public ServletAndFilterConfig(ServletContext servletContext,
			String name, Map<String, String> initParamMap) {
		this.servletContext = Objects.nonNull(servletContext);
		this.name = Objects.nonNull(name);
		this.initParamMap = new HashMap<String, String>(initParamMap);
	}

	public ServletContext getServletContext() {
		return servletContext;
	}

	public String getServletName() {
		return name;
	}

	public String getFilterName() {
		return name;
	}

	public Enumeration<String> getInitParameterNames() {
		return Collections.enumeration(initParamMap.keySet());
	}

	public String getInitParameter(String name) {
		return initParamMap.get(name);
	}
	
	/**
	 * Wraps {@code filterConfig} and overrides init parameter.
	 * 
	 * @param filterConfig
	 *            filter config to be wrapped.
	 * @param initParamMap
	 *            init parameters to override.
	 * @return wrapped filter config.
	 */
	public static FilterConfig overrideInitParameters(FilterConfig filterConfig,
			Map<String, String> initParamMap) {
		
		Map<String, String> overriddenParamMap = new TreeMap<String, String>();

		@SuppressWarnings("unchecked")
		Enumeration<String> oriParamNameEnm = filterConfig.getInitParameterNames();
		
		while (oriParamNameEnm.hasMoreElements()) {
			String name = oriParamNameEnm.nextElement();
			String value = filterConfig.getInitParameter(name);
			
			overriddenParamMap.put(name, value);
		}
		overriddenParamMap.putAll(initParamMap);
		return new ServletAndFilterConfig(filterConfig.getServletContext(),
				filterConfig.getFilterName(), overriddenParamMap);
	}

}
