/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package org.apithefire.servlet.embedded;

import java.util.Comparator;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.SortedMap;
import java.util.TreeMap;

import javax.servlet.Servlet;

import org.apithefire.util.lang.Objects;
import org.apithefire.util.lang.Path;

class ServletPathMap {
	
	public class Match {

		private Servlet servlet;
		private Path servletPath;
		private Path pathInfo;
		
		protected Match(Servlet servlet, Path servletPath, Path pathInfo) {
			this.servlet = Objects.nonNull(servlet);
			this.servletPath = Objects.nonNull(servletPath);
			this.pathInfo = Objects.nonNull(pathInfo);
		}
		
		public Servlet getServlet() {
			return servlet;
		}
		
		public Path getServletPath() {
			return servletPath;
		}
		
		public Path getPathInfo() {
			return pathInfo;
		}
		
	}

	private static class LengthFirstPathComparator implements Comparator<Path> {

		public int compare(Path path1, Path path2) {
			int diff = path2.length() - path1.length();

			if (diff != 0) {
				return diff;
			}
			return path2.compareTo(path1);
		}

	}

	private static final Comparator<Path> LENGTH_FIRST_PATH_COMPARATOR =
		new LengthFirstPathComparator();
	
	private final SortedMap<Path, Servlet> exactServletMap;
	private final SortedMap<Path, Servlet> prefixServletMap;
	private final SortedMap<Path, Servlet> suffixServletMap;
	private Servlet defaultServlet;

	public ServletPathMap() {
		exactServletMap = new TreeMap<Path, Servlet>();
		prefixServletMap = new TreeMap<Path, Servlet>(LENGTH_FIRST_PATH_COMPARATOR);
		suffixServletMap = new TreeMap<Path, Servlet>(LENGTH_FIRST_PATH_COMPARATOR);
	}
	
	public Match getMatch(Path path) {
		Servlet servlet = exactServletMap.get(Objects.nonNull(path));
		
		if (servlet != null) {
			return new Match(servlet, path, Path.EMPTY_PATH);
		}
		
		Match Match = getPrefixMatch(path);
		
		if (Match != null) {
			return Match;
		}
		
		Match = getSuffixMatch(path);
		
		if (Match != null) {
			return Match;
		}
		if (defaultServlet == null) {
			return null;
		}
		return new Match(defaultServlet, path, Path.EMPTY_PATH);
	}
	
	private Match getPrefixMatch(Path path) {
		if (prefixServletMap.isEmpty()) {
			return null;
		}
		
		Map<Path, Servlet> tailMap = prefixServletMap.tailMap(path);
		
		if (tailMap.isEmpty()) {
			return null;
		}
		
		Map.Entry<Path, Servlet> firstMatch = tailMap.entrySet().iterator().next();
		Path prefix = firstMatch.getKey();
		Servlet servlet = firstMatch.getValue();
		
		if (!path.startsWith(prefix)) {
			return null;
		}
		
		Path servletPath = path.subPath(0, prefix.length());
		Path pathInfoPath = path.subPath(prefix.length());
		
		return new Match(servlet, servletPath, pathInfoPath);
	}
	
	private Match getSuffixMatch(Path path) {
		Path suffix = EmbeddedUtil.splitLastComponentAsSuffix(path);
		
		if (suffix == null) {
			return null;
		}
		
		Map<Path, Servlet> tailMap = suffixServletMap.tailMap(suffix);
		
		if (tailMap.isEmpty()) {
			return null;
		}
		
		Map.Entry<Path, Servlet> firstMatch = tailMap.entrySet().iterator().next();
		Path matchSuffix = firstMatch.getKey();
		Servlet servlet = firstMatch.getValue();
		
		if (!suffix.endsWith(matchSuffix)) {
			return null;
		}
		return new Match(servlet, path, Path.EMPTY_PATH);
	}
	
	public void mapExact(Path exactPath, Servlet servlet) {
		exactServletMap.put(Objects.nonNull(exactPath), Objects.nonNull(servlet));
	}
	
	public void mapPrefix(Path prefix, Servlet servlet) {
		prefixServletMap.put(Objects.nonNull(prefix), Objects.nonNull(servlet));
	}
	
	public void mapSuffix(Path suffix, Servlet servlet) {
		suffixServletMap.put(Objects.nonNull(suffix), Objects.nonNull(servlet));
	}
	
	public void mapDefault(Servlet servlet) {
		defaultServlet = servlet;
	}
	
	public Set<Servlet> getServlets() {
		Set<Servlet> servlets = new HashSet<Servlet>();
		
		servlets.addAll(exactServletMap.values());
		servlets.addAll(prefixServletMap.values());
		servlets.addAll(suffixServletMap.values());
		
		if (defaultServlet != null) {
			servlets.add(defaultServlet);
		}
		return servlets;
	}
	
	@Override
	public ServletPathMap clone() {
		ServletPathMap clone = new ServletPathMap();
		
		clone.exactServletMap.putAll(exactServletMap);
		clone.prefixServletMap.putAll(prefixServletMap);
		clone.suffixServletMap.putAll(suffixServletMap);
		clone.defaultServlet = defaultServlet;
		return clone;
	}

}
